/*
 * S/MIME Message Parser, v0.1 (alpha)
 *
 * Copyright 1997, Chris Hall - Counterpane Systems
 *
 * This is a partially implemented S/MIME message parser.  It parses the IV
 * and ciphertext from a message containing EnvelopedData or
 * SignedAndEnvelopedData.  This code was implemented as per the Internet
 * Draft draft-dusse-smime-msg-02.txt, pkcs-7, and the ASN.1 description
 * written by B. Kaliski.  These are available as:
 *
 *   http://www.rsa.com/pub/pkcs/ascii/pkcs-7.asc
 *   ftp://ds.internic.net/internet-drafts/draft-dusse-smime-msg-02.txt
 *   http://www.rsa.com/pub/pkcs/ascii/layman.asc
 *
 * The current version of the code cannot handle the following:
 *
 * - A top-level Data type.  As I read the spec, this shouldn't ever happen
 *    so it probably isn't a problem.
 * - ExtendedCertificatesAndCertificates, CertificateRevocationLists,
 *    SingerInfos, IssuerAndSerialNumber, and SignedData types.  These are
 *    necessary to extract the ciphertext.
 *
 * To use the program one must simply provide a binary S/MIME message.  The
 * exact command-line syntax is:
 *
 *   parse <input file> <output file>
 *
 * The resulting output file will contain the 8-byte IV followed by the
 * ciphertext.  Note, this is only true if the program is able to find data
 * encrypted with RC2/40.  Otherwise the output file will be empty (or may
 * contain an IV).
 *
 * In addition to placing the IV and ciphertext in the output file, the
 * program will print out the names of various components of the file as
 * it encounters them.  To turn this off, change the #define VERBOSE line
 * to '#define VERBOSE 0' and recompile the program.
 *
 * I have tried hard to check for errors wherever possible in order to make
 * this program as robust as possible.  However, I could have screwed up
 * so I would appreciate information about any bugs.
 *
 * Finally, I have made little effort to comment the code.  Most of it
 * should be self-explanatory as the majority of the code is spent in
 * parsing DER-encoded data and checking for errors.  There is almost
 * certainly a more efficient approach than the one I took here, but it
 * would only have been efficient as far as code length and speed goes.
 * I had to learn ASN.1 in order to write this program and did not have the
 * time to write a proper parser.  My guess is that many ASN.1 to C parsers
 * simply produce something similar to this code anyway.  So please don't
 * waste your time nor mine by sending me e-mail telling me how to make the
 * code more efficient.  *I know it isn't efficient*.  However, it works!
 */

#include <stdio.h>

#define VERBOSE 1

#define ASSERT(e) \
do { \
    if (!(e)) { \
	fprintf(stderr, "assert failed on line %d.\n", __LINE__); \
	exit(-1); \
    } \
} while (0)

typedef enum error_t {
    ERROR_NONE = 0,
    ERROR_BAD_INTEGER,
    ERROR_BAD_LENGTH,
    ERROR_INDEFINITE_LENGTH,
    ERROR_IO_ERROR,
    ERROR_INVALID_OBJECT_IDENTIFIER,
    ERROR_MISSING_DATA,
    ERROR_TOO_MUCH_DATA,
    ERROR_UNEXPECTED_CONTENT_TYPE,
    ERROR_UNEXPECTED_OBJECT_IDENTIFIER,
    ERROR_UNEXPECTED_TYPE,
    ERROR_UNEXPECTED_VERSION,
    ERROR_UNKNOWN_ALGORITHM,
    ERROR_UNKNOWN_CONTENT_TYPE,
    ERROR_UNKNOWN_DIGEST_ALGORITHM,
    ERROR_UNKNOWN_TYPE,
    ERROR_NUM_ERRORS
} error_t, *error_p;

typedef enum type_t {
    TYPE_UNKNOWN = -1,
    TYPE_NONE,
    TYPE_INTEGER,
    TYPE_BIT_STRING,
    TYPE_OCTET_STRING,
    TYPE_NULL,
    TYPE_OBJECT_IDENTIFIER,
    TYPE_SEQUENCE,
    TYPE_SET,
    TYPE_PrintableString,
    TYPE_T61String,
    TYPE_TA5String,
    TYPE_UTCTime,
    TYPE_APPLICATION_0,
    TYPE_APPLICATION_1,
    TYPE_CONTEXT_SPECIFIC_0,
    TYPE_CONTEXT_SPECIFIC_1,
    TYPE_PRIVATE_0,
    TYPE_PRIVATE_1,
    TYPE_NUM_TYPES
} type_t, *type_p;

typedef enum content_type_t {
    CONTENT_TYPE_UNKNOWN = -1,
    CONTENT_TYPE_DATA,
    CONTENT_TYPE_SIGNED_DATA,
    CONTENT_TYPE_ENVELOPED_DATA,
    CONTENT_TYPE_SIGNED_AND_ENVELOPED_DATA,
    CONTENT_TYPE_DIGESTED_DATA,
    CONTENT_TYPE_ENCRYPTED_DATA,
    CONTENT_TYPE_NUM_TYPES
} content_type_t, *content_type_p;

typedef enum algorithm_t {
    ALGORITHM_UNKNOWN = -1,
    ALGORITHM_RSA,
    ALGORITHM_RC2_40_CBC,
    ALGORITHM_RC2_64_CBC,
    ALGORITHM_RC2_128_CBC,
    ALGORITHM_NUM_ALGORITHMS
} algorithm_t, *algorithm_p;

typedef enum digest_algorithm_t {
    DIGEST_ALGORITHM_UNKNOWN = -1,
    DIGEST_ALGORITHM_SHA1,
    DIGEST_ALGORITHM_MD5,
    DIGEST_ALGORITHM_NUM_ALGORITHMS
} digest_algorithm_t, *digest_algorithm_p;

algorithm_t encryption_algorithm;
char	spaces[] = "                    ", *indent = spaces + sizeof(spaces);
FILE	*output;

error_t read_ContentInfo(FILE *f, long *length_limit);

void print_error(error_t error)
{
    char    *msg;

    switch (error) {
	case ERROR_NONE: msg = "no error"; break;
	case ERROR_BAD_INTEGER:	msg = "bad integer"; break;
	case ERROR_BAD_LENGTH: msg = "bad length encountered"; break;
	case ERROR_INDEFINITE_LENGTH:
	    msg = "unexpected indefinite length"; break;
	case ERROR_INVALID_OBJECT_IDENTIFIER:
	    msg = "invalid object identifier"; break;
	case ERROR_IO_ERROR: msg = "error reading from file"; break;
	case ERROR_MISSING_DATA: msg = "missing data"; break;
	case ERROR_TOO_MUCH_DATA: msg = "too much data"; break;
	case ERROR_UNEXPECTED_CONTENT_TYPE:
	    msg = "unexpected content type"; break;
	case ERROR_UNEXPECTED_OBJECT_IDENTIFIER:
	    msg = "unexpected object identifier"; break;
	case ERROR_UNEXPECTED_TYPE: msg = "unexpected ASN.1 type"; break;
	case ERROR_UNEXPECTED_VERSION:
	    msg = "unexpected version number"; break;
	case ERROR_UNKNOWN_ALGORITHM:
	    msg = "unknown algorithm"; break;
	case ERROR_UNKNOWN_CONTENT_TYPE:
	    msg = "unknown content type"; break;
	case ERROR_UNKNOWN_DIGEST_ALGORITHM:
	    msg = "unknown digest algorithm"; break;
	case ERROR_UNKNOWN_TYPE: msg = "unknown ASN.1 type"; break;
	default: msg = "unknown error";	break;
    }
    fprintf(stderr, "%s.\n", msg);
}

void usage(char *progname)
{
    fprintf(stderr, "%s: <input filename> <output filename>\n");
}

/*
 * read_length() - reads a proper length from the data stream.
 *
 * input:
 *   f - stream to read data from.
 *   length_limit - maximum number of octets we can read in order to
 *     determine the length.
 *   indefinite - whether or not indefinite lengths are allowed.  NULL
 *     if indefinite lengths are not allowed.
 *
 * output:
 *   length - length as read from data stream.
 *   indefinite - indicates whether the length was indefinite.
 */

error_t read_length(FILE *f, long *length_limit, long *length, short *indefinite)
{
    unsigned char   b, l;

    if (*length_limit == 0)
	return ERROR_MISSING_DATA;

    if (fread(&b, 1, 1, f) != 1)
	return ERROR_IO_ERROR;
    (*length_limit)--;

    if (b == 0x80) {
	if (indefinite == NULL)
	    return ERROR_INDEFINITE_LENGTH;
	*length = *length_limit;
	*indefinite = 1;
	return ERROR_NONE;
    } else if (indefinite != NULL)
	*indefinite = 0;

    if ((b & 0x80) == 0) {
	*length = b;
	return ERROR_NONE;
    }

    /* need to read the long encoding of the length */
    b &= 0x7F;
    if (b > *length_limit)
	return ERROR_MISSING_DATA;

    *length = 0;
    while (b--) {
	(*length_limit)--;
	if (fread(&l, 1, 1, f) != 1)
	    return ERROR_IO_ERROR;

	*length = (*length)*256 + l;
    }
    return ERROR_NONE;
}

error_t read_type(FILE *f, long *length_limit, type_t *type)
{
    unsigned char   b;

    if ((*length_limit)-- < 1)
	return ERROR_MISSING_DATA;
    if (fread(&b, 1, 1, f) != 1)
	return ERROR_IO_ERROR;

    if ((b & 0xC0) == 0) {
	switch (b & 0x3F) {
	    case  0: *type = TYPE_NONE;	    	    	break;
	    case  2: *type = TYPE_INTEGER;  	    	break;
	    case  3: *type = TYPE_BIT_STRING;	    	break;
	    case  4: *type = TYPE_OCTET_STRING;	    	break;
	    case  5: *type = TYPE_NULL;	    	    	break;
	    case  6: *type = TYPE_OBJECT_IDENTIFIER;	break;
	    case 48: *type = TYPE_SEQUENCE; 	    	break;
	    case 49: *type = TYPE_SET;	    	    	break;
	    case 19: *type = TYPE_PrintableString;  	break;
	    case 20: *type = TYPE_T61String;	    	break;
	    case 22: *type = TYPE_TA5String;	    	break;
	    case 23: *type = TYPE_UTCTime;  	    	break;
	    default: return ERROR_UNKNOWN_TYPE;
	}
	return ERROR_NONE;
    }

    if ((b & 0x1F) > (TYPE_CONTEXT_SPECIFIC_0 - TYPE_APPLICATION_0))
	return ERROR_UNKNOWN_TYPE;  /* not really unknown, but... */
    switch (b >> 6) {
	case 1:
	    *type = TYPE_APPLICATION_0 + (b & 0x1F);
	    ASSERT(*type < TYPE_CONTEXT_SPECIFIC_0);
	    break;

	case 2:
	    *type = TYPE_CONTEXT_SPECIFIC_0 + (b & 0x1F);
	    ASSERT(*type < TYPE_PRIVATE_0);
	    break;

	case 3:
	    *type = TYPE_PRIVATE_0 + (b & 0x1F);
	    ASSERT(*type < TYPE_NUM_TYPES);
	    break;
    }
    return ERROR_NONE;
}

error_t read_INTEGER(FILE *f, long *length_limit, int *i)
{
    unsigned char   b;
    char    first = 1;
    int	    sign;
    long    length;
    error_t result;

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

    *length_limit -= length;
    if (length+1 > sizeof(int))
	return ERROR_BAD_INTEGER;

    *i = 0;
    while (length--) {
	if (fread(&b, 1, 1, f) != 1)
	    return ERROR_IO_ERROR;

	if (first) {
	    sign = (b & 0x40) ? -1 : 1;
	    first = 0;
	}

	*i = (*i)*256 + b;
    }
    *i *= sign;

    return ERROR_NONE;
}

error_t read_OCTET_STRING(FILE *f, long *length_limit, unsigned char *string,
    short *len)
{
    error_t result;
    type_t  type;
    long    length;
    unsigned char   b, *c;
    short   buf_len, l, i = 1;  /* allow an indefinite length */

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

    *length_limit -= length;

    if (i == 0) {
	if (string != NULL && length > *len)
	    return ERROR_BAD_LENGTH;

	if (string == NULL) {
	    if (fseek(f, length, SEEK_CUR) != 0)
		return ERROR_IO_ERROR;
	    return ERROR_NONE;
	}
	if (fread(string, 1, length, f) != length)
	    return ERROR_IO_ERROR;
	*len = length;
    } else {
	c = string;
	buf_len = *len;
	*len = 0;
	while (length > 0) {
	    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
		return result;
	    if (type != TYPE_OCTET_STRING && type != TYPE_NONE)
		return ERROR_UNEXPECTED_TYPE;
	    if (type == TYPE_NONE) {
		*length_limit += length;
		if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
		    return result;
		if (length != 0)
		    return ERROR_BAD_LENGTH;
		return ERROR_NONE;
	    }
	    l = buf_len;
	    if ((result = read_OCTET_STRING(f, &length, c, &l)) != ERROR_NONE)
		return result;
	    if (c != NULL) {
		c += l;
		*len += l;
		buf_len -= l;
	    }
	}
	return ERROR_BAD_LENGTH;
    }

    return ERROR_NONE;
}

error_t read_OBJECT_IDENTIFIER(FILE *f, long *length_limit,
    unsigned char *buffer, short *len)
{
    error_t result;
    type_t  type;
    long    length;

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;
    if (length > *len)
	return ERROR_BAD_LENGTH;

    *len = length;
    *length_limit -= length;
    if (fread(buffer, 1, length, f) != length)
	return ERROR_IO_ERROR;

    /* error-check the object identifier */
    if (length > 1 && (buffer[length-1] & 0x80) != 0)
	return ERROR_INVALID_OBJECT_IDENTIFIER;

    return ERROR_NONE;
}

error_t read_NULL(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit || length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

/*
 * UNIMPLEMENTED!!!
 */

error_t read_ExtendedCertificatesAndCertificates(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sExtendedCertificatesAndCertificates (SKIPPED)\n", indent);
#endif

    /* skip the data so we don't have to implement this right now */
    if (fseek(f, length, SEEK_CUR) != 0)
	return ERROR_IO_ERROR;
    *length_limit -= length;

    return ERROR_NONE;
}

/*
 * UNIMPLEMENTED!!!
 */

error_t read_CertificateRevocationLists(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sCertificateRevocationLists (SKIPPED)\n", indent);
#endif

    /* skip the data so we don't have to implement this right now */
    if (fseek(f, length, SEEK_CUR) != 0)
	return ERROR_IO_ERROR;
    *length_limit -= length;

    return ERROR_NONE;
}

/*
 * UNIMPLEMENTED!!!
 */

error_t read_SingerInfos(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sSignerInfos (SKIPPED)\n", indent);
#endif

    /* skip the data so we don't have to implement this right now */
    if (fseek(f, length, SEEK_CUR) != 0)
	return ERROR_IO_ERROR;
    *length_limit -= length;

    return ERROR_NONE;
}

/*
 * UNIMPLEMENTED!!!
 */

error_t read_IssuerAndSerialNumber(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sIssuerAndSerialNumber (SKIPPED)\n", indent);
#endif

    /* skip the data so we don't have to implement this right now */
    if (fseek(f, length, SEEK_CUR) != 0)
	return ERROR_IO_ERROR;
    *length_limit -= length;

    return ERROR_NONE;
}

error_t read_ContentType(FILE *f, long *length_limit, content_type_t *type)
{
    /* sig[] = { iso(1) member-body(2) US(840) rsadsi(113549) pkcs(1) 7 }
     * RSA Data Security Inc., PKCS
     */
    unsigned char   sig[] = {0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x07};
    unsigned char   buffer[9];
    error_t result;
    long    length;
    short   id_len = sizeof(buffer);

    ASSERT(sizeof(sig) == sizeof(buffer)-1);

#if VERBOSE == 1
    printf("%sContentType (OBJECT_IDENTIFIER)\n", indent);
#endif

    /* ContentType ::= OBJECT_IDENTIFIER */
    if ((result = read_OBJECT_IDENTIFIER(f, length_limit, buffer, &id_len)) != ERROR_NONE)
	return result;
    if (id_len != sizeof(buffer))
	return ERROR_UNKNOWN_CONTENT_TYPE;

    /* parse it */
    if (memcmp(buffer, sig, sizeof(sig)) != 0)
	return ERROR_UNKNOWN_CONTENT_TYPE;
    switch (buffer[sizeof(buffer)-1]) {
	case 1: *type = CONTENT_TYPE_DATA;  break;
	case 2: *type = CONTENT_TYPE_SIGNED_DATA;   break;
	case 3: *type = CONTENT_TYPE_ENVELOPED_DATA;	break;
	case 4: *type = CONTENT_TYPE_SIGNED_AND_ENVELOPED_DATA;	break;
	case 5: *type = CONTENT_TYPE_DIGESTED_DATA; break;
	case 6: *type = CONTENT_TYPE_ENCRYPTED_DATA;	break;
	default:    return ERROR_UNKNOWN_CONTENT_TYPE;
    }
    return ERROR_NONE;
}

error_t read_Data(FILE *f, long *length_limit, unsigned char *output)
{
    return ERROR_NONE;
}

/*
 * UNIMPLEMENTED!!!
 */

error_t read_data(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sData (SKIPPED)\n", indent);
#endif

    /* skip the data so we don't have to implement this right now */
    if (fseek(f, length, SEEK_CUR) != 0)
	return ERROR_IO_ERROR;
    *length_limit -= length;

    return ERROR_NONE;
}

error_t read_SignedData(FILE *f, long *length_limit)
{
    return ERROR_NONE;
}

/*
 * UNIMPLEMENTED!!!
 */

error_t read_signedData(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sSignedData (SKIPPED)\n", indent);
#endif

    /* skip the data so we don't have to implement this right now */
    if (fseek(f, length, SEEK_CUR) != 0)
	return ERROR_IO_ERROR;
    *length_limit -= length;

    return ERROR_NONE;
}

error_t read_EncryptedContent(FILE *f, long *length_limit)
{
    unsigned char   *ciphertext;
    error_t result;
    type_t  type;
    short   c_len, j, i = 1;	/* allow an indefinite length */

#if VERBOSE == 1
    printf("%sEncryptedContent (OCTET_STRING)\n", indent);
#endif
    if ((ciphertext = (unsigned char*)malloc(*length_limit)) == NULL) {
	fprintf(stderr, "error allocating memory for ciphertext.\n");
	return ERROR_NONE;
    }

    c_len = *length_limit;
    if ((result = read_OCTET_STRING(f, length_limit, ciphertext, &c_len)) != ERROR_NONE)
	return result;

#if VERBOSE == 1
    printf("%s", indent-2);
    for (j = 0; j < c_len; j++) {
	printf("%02x ", ciphertext[j]);
	if (j % 24 == 23)
	    printf("\n%s", indent-2);
    }
    printf("\n");
#endif
    fwrite(ciphertext, 1, c_len, output);

    free(ciphertext);

    return ERROR_NONE;
}

error_t read_KeyEncryptionAlgorithmIdentifier(FILE *f, long *length_limit,
    algorithm_t *algorithm)
{
    /* sig[] = { iso(1) member-body(2) US(840) rsadsi(113549) pkcs(1) pkcs-1(1) }
     * RSA Data Security Inc., PKCS
     */
    unsigned char   sig[] = {0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01};
    unsigned char   buffer[9];
    error_t result;
    type_t  type;
    long    length;
    short   id_len = sizeof(buffer);

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sKeyEncryptionAlgorithmIdentifier (OBJECT_IDENTIFIER)\n", indent);
#endif
    *length_limit -= length;

    /* read an OBJECT IDENTIFIER */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_OBJECT_IDENTIFIER)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_OBJECT_IDENTIFIER(f, &length, buffer, &id_len)) != ERROR_NONE)
	return result;
    if (id_len != sizeof(buffer))
	return ERROR_UNKNOWN_ALGORITHM;

    /* parse it */
    if (memcmp(buffer, sig, sizeof(sig)) != 0)
	return ERROR_UNKNOWN_ALGORITHM;
    switch (buffer[sizeof(buffer)-1]) {
	case 1:
	    *algorithm = ALGORITHM_RSA;

	    /* read the parameter */
	    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
		return result;
	    if (type != TYPE_NULL)
		return ERROR_UNEXPECTED_TYPE;
#if VERBOSE == 1
	    printf("%sRsaParameter (NULL)\n", indent-1);
#endif
	    if ((result = read_NULL(f, &length)) != ERROR_NONE)
		return result;
	    break;

	default:    return ERROR_UNKNOWN_ALGORITHM;
    }

    if (length != 0)
	return ERROR_TOO_MUCH_DATA;

    return ERROR_NONE;
}

error_t read_rc2Parameter(FILE *f, long *length_limit, int *version)
{
    unsigned char   iv[8];
    error_t result;
    type_t  type;
    long    length;
    short   j, iv_len = 8, i = 1;  /* allow an indefinite length */

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sRC2-CBC parameter (SEQUENCE)\n", indent);
#endif
    *length_limit -= length;

    /* read rc2ParameterVersion */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_INTEGER)
	return ERROR_UNEXPECTED_TYPE;
#if VERBOSE == 1
    printf("%src2ParameterVersion (INTEGER)\n", indent-1);
#endif
    if ((result = read_INTEGER(f, &length, version)) != ERROR_NONE)
	return result;

    /* read the IV */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_OCTET_STRING)
	return ERROR_UNEXPECTED_TYPE;
    if (length > 9)
	return ERROR_TOO_MUCH_DATA;
#if VERBOSE == 1
    printf("%sIV (OCTET_STRING)\n", indent-1);
#endif
    if ((result = read_OCTET_STRING(f, &length, iv, &iv_len)) != ERROR_NONE)
	return result;

#if VERBOSE == 1
    printf("%s", indent-3);
    for (j = 0; j < iv_len; j++)
	printf("%02x ", iv[j]);
    printf("\n");
#endif
    fwrite(iv, 1, iv_len, output);

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_ContentEncryptionAlgorithmIdentifier(FILE *f, long *length_limit,
    algorithm_t *algorithm)
{
    /* sig[] = { iso(1) member-body(2) US(840) rsadsi(113549) pkcs(1)
     *           encryptionAlgorithm(3) }
     * RSA Data Security Inc., PKCS
     */
    unsigned char   sig[] = {0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x03};
    unsigned char   buffer[8];
    error_t result;
    type_t  type;
    long    length;
    int	    version;
    short   id_len = sizeof(buffer);

    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sContentEncryptionAlgorithmIdentifier (OBJECT_IDENTIFIER)\n", indent);
#endif
    *length_limit -= length;

    /* read an OBJECT IDENTIFIER */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_OBJECT_IDENTIFIER)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_OBJECT_IDENTIFIER(f, &length, buffer, &id_len)) != ERROR_NONE)
	return result;
    if (id_len != sizeof(buffer))
	return ERROR_UNKNOWN_ALGORITHM;

    /* parse it */
    if (memcmp(buffer, sig, sizeof(sig)) != 0)
	return ERROR_UNKNOWN_ALGORITHM;
    switch (buffer[sizeof(buffer)-1]) {
	case 2:
	    /* RC2-CBC */

	    /* read the parameter */
	    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
		return result;
	    if (type != TYPE_SEQUENCE)
		return ERROR_UNEXPECTED_TYPE;
	    indent--;
	    if ((result = read_rc2Parameter(f, &length, &version)) != ERROR_NONE)
		return result;
	    indent++;

	    /* determine which algorithm it is */
	    switch (version) {
		case 160:
#if VERBOSE == 1
		    printf("%s(data encrypted with RC2/40)\n", indent);
#endif
		    *algorithm = ALGORITHM_RC2_40_CBC;
		    break;
		case 120:
#if VERBOSE == 1
		    printf("%s(data encrypted with RC2/64)\n", indent);
#endif
		    *algorithm = ALGORITHM_RC2_64_CBC;
		    break;
		case  58:
#if VERBOSE == 1
		    printf("%s(data encrypted with RC2/128)\n", indent);
#endif
		    *algorithm = ALGORITHM_RC2_128_CBC;
		    break;
		default: return ERROR_UNKNOWN_ALGORITHM;
	    }

	    if (length != 0)
		return ERROR_MISSING_DATA;
	    break;

	case 7:
	    /* DES-EDE3-CBC, don't want to handle right now */
	    return ERROR_UNKNOWN_ALGORITHM;
	    break;

	default:    return ERROR_UNKNOWN_ALGORITHM;
    }

    if (length != 0)
	return ERROR_MISSING_DATA;

    return ERROR_NONE;
}

error_t read_RecipientInfo(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    int	    version;
    short   i = 1;  /* allow an indefinite length */
    algorithm_t	algorithm;

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sRecipientInfo (SEQUENCE)\n", indent);
#endif
    *length_limit -= length;

    /* read version */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_INTEGER)
	return ERROR_UNEXPECTED_TYPE;
#if VERBOSE == 1
    printf("%sVersion (INTEGER)\n", indent-1);
#endif
    if ((result = read_INTEGER(f, &length, &version)) != ERROR_NONE)
	return result;
    if (version != 0)
	return ERROR_UNEXPECTED_VERSION;

    /* read issuerAndSerialNumber */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    indent--;
    if ((result = read_IssuerAndSerialNumber(f, &length)) != ERROR_NONE)
	return result;

    /* read keyEncryptionAlgorithmIdentifier */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_KeyEncryptionAlgorithmIdentifier(f, &length, &algorithm)) != ERROR_NONE)
	return result;
#if VERBOSE == 1
    switch (algorithm) {
	case ALGORITHM_RSA:
	    printf("%s(key encrypted with RSA)\n", indent); break;
	default:
	    printf("%s(unknown key encryption algorithm)\n", indent); break;
    }
#endif

    /* read encryptedKey */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_OCTET_STRING)
	return ERROR_UNEXPECTED_TYPE;
#if VERBOSE == 1
    printf("%sEncryptedKey (OCTET_STRING)\n", indent);
#endif
    if ((result = read_OCTET_STRING(f, &length, NULL, NULL)) != ERROR_NONE)
	return result;
    indent++;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_RecipientInfos(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    short   i = 1;  	/* allow an indefinite length */

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sRecipientInfos (SET OF RecipientInfo)\n", indent);
#endif
    *length_limit -= length;

    while (length) {
	/* read recipientInfo */
	if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	    return result;
	if (type != TYPE_SEQUENCE && type != TYPE_NONE)
	    return ERROR_UNEXPECTED_TYPE;
	if (i != 0 && type == TYPE_NONE) {
	    long    l;
	    if ((result = read_length(f, &length, &l, NULL)) != ERROR_NONE)
		return result;
	    *length_limit += length;
	    if (l != 0)
		return ERROR_BAD_LENGTH;
	    return ERROR_NONE;
	}
	indent--;
	if ((result = read_RecipientInfo(f, &length)) != ERROR_NONE)
	    return result;
	indent++;
    }

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    ASSERT(0);	/* shouldn't get here */

    return ERROR_NONE;
}

error_t read_EncryptedContentInfo(FILE *f, long *length_limit)
{
    content_type_t  content_type;
    error_t result;
    type_t  type;
    long    length;
    short   i = 1;  /* allow an indefinite length */

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sEncryptedContentInfo (SEQUENCE)\n", indent);
#endif
    *length_limit -= length;

    /* read contentType */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_OBJECT_IDENTIFIER)
	return ERROR_UNEXPECTED_TYPE;
    indent--;
    if ((result = read_ContentType(f, &length, &content_type)) != ERROR_NONE)
	return result;
    if (content_type != CONTENT_TYPE_DATA)
	return ERROR_UNEXPECTED_CONTENT_TYPE;

    /* read contentEncryptionAlgorithm */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_ContentEncryptionAlgorithmIdentifier(f, &length,
	    &encryption_algorithm)) != ERROR_NONE)
	return result;

    /* read encryptedContent */
    if (length == 0)
	return ERROR_NONE;  /* no encrypted data, ;( */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_CONTEXT_SPECIFIC_0)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_EncryptedContent(f, &length)) != ERROR_NONE)
	return result;
    indent++;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_EnvelopedData(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    int	    version;
    short   i = 1;  /* allow an indefinite length */

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sEnvelopedData (SEQUENCE)\n", indent);
#endif
    *length_limit -= length;

    /* read version */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_INTEGER)
	return ERROR_UNEXPECTED_TYPE;
#if VERBOSE == 1
    printf("%sVersion (INTEGER)\n", indent-1);
#endif
    if ((result = read_INTEGER(f, &length, &version)) != ERROR_NONE)
	return result;
    if (version != 0)
	return ERROR_UNEXPECTED_VERSION;

    /* read recipientInfos */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SET)
	return ERROR_UNEXPECTED_TYPE;
    indent--;
    if ((result = read_RecipientInfos(f, &length)) != ERROR_NONE)
	return result;

    /* read encryptedContentInfo */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_EncryptedContentInfo(f, &length)) != ERROR_NONE)
	return result;
    indent++;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_envelopedData(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    short   i = 1;  /* allow an indefinite length */

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

    *length_limit -= length;

    /* read envelopedData */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_EnvelopedData(f, &length)) != ERROR_NONE)
	return result;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_DigestAlgorithmIdentifier(FILE *f, long *length_limit,
    digest_algorithm_t *digest)
{
    unsigned char   md5_sig[] = {0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x02, 0x05};
    unsigned char   sha_sig[] = {0x2b, 0x0e, 0x03, 0x02, 0x1a};
    unsigned char   buffer[sizeof(md5_sig)];
    error_t result;
    type_t  type;
    long    length;
    short   id_len = sizeof(buffer), i = 1;  /* allow an indefinite length */

    ASSERT(sizeof(sha_sig) <= sizeof(buffer) && sizeof(md5_sig) <= sizeof(buffer));

#if VERBOSE == 1
    printf("%sDigestAlgorithmIdentifier (OBJECT_IDENTIFIER)\n", indent);
#endif
    if ((result = read_OBJECT_IDENTIFIER(f, length_limit, buffer, &id_len)) != ERROR_NONE)
	return result;

    if (id_len == sizeof(sha_sig) && memcmp(sha_sig, buffer, id_len) == 0)
	*digest = DIGEST_ALGORITHM_SHA1;
    else if (id_len == sizeof(md5_sig) && memcmp(md5_sig, buffer, id_len) == 0)
	*digest = DIGEST_ALGORITHM_MD5;
    else
	return ERROR_UNKNOWN_DIGEST_ALGORITHM;

    return ERROR_NONE;
}

error_t read_DigestAlgorithmIdentifiers(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    short   i = 1;  /* allow an indefinite length */
    digest_algorithm_t	digest;

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sDigestAlgorithmIdentifiers (SET OF DigestAlgorithmIdentifier)\n", indent);
#endif
    *length_limit -= length;

    while (length) {
	if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	    return result;
	if (type != TYPE_OBJECT_IDENTIFIER && type != TYPE_NONE)
	    return ERROR_UNEXPECTED_TYPE;
	if (type == TYPE_NONE) {
	    if (i == 0)
		return ERROR_UNEXPECTED_TYPE;
	    *length_limit += length;
	    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
		return result;
	    if (length != 0)
		return ERROR_BAD_LENGTH;
	    return ERROR_NONE;
	}
	indent--;
	if ((result = read_DigestAlgorithmIdentifier(f, &length, &digest)) != ERROR_NONE)
	    return result;
#if VERBOSE == 1
	switch (digest) {
	    case DIGEST_ALGORITHM_SHA1:
		printf("%s(data hashed with SHA-1)\n", indent); break;
	    case DIGEST_ALGORITHM_MD5:
		printf("%s(data hashed with MD5)\n", indent); break;
	    default:
		printf("%s(unknown hashing algorithm)\n", indent); break;
	}
#endif
	indent++;
    }

    if (i != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_SignedAndEnvelopedData(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    short   i = 1;  /* allow an indefinite length */
    int	    version;

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sSignedAndEnvelopedData (SEQUENCE)\n", indent);
#endif
    *length_limit -= length;

    /* read version */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_INTEGER)
	return ERROR_UNEXPECTED_TYPE;
#if VERBOSE == 1
    printf("%sVersion (INTEGER)\n", indent-1);
#endif
    if ((result = read_INTEGER(f, &length, &version)) != ERROR_NONE)
	return result;
    if (version != 0)
	return ERROR_UNEXPECTED_VERSION;

    /* read recipientInfos */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SET)
	return ERROR_UNEXPECTED_TYPE;
    indent--;
    if ((result = read_RecipientInfos(f, &length)) != ERROR_NONE)
	return result;

    /* read digestAlgorithmIdentifiers */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SET)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_DigestAlgorithmIdentifiers(f, &length)) != ERROR_NONE)
	return result;

    /* read encryptedContentInfo */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_EncryptedContentInfo(f, &length)) != ERROR_NONE)
	return result;

    /* try to read certificates */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_CONTEXT_SPECIFIC_0 && type != TYPE_CONTEXT_SPECIFIC_1 &&
	type != TYPE_SET)
	return ERROR_UNEXPECTED_TYPE;

    if (type == TYPE_CONTEXT_SPECIFIC_0) {
	if ((result = read_ExtendedCertificatesAndCertificates(f, &length)) != ERROR_NONE)
	    return result;

	if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	    return result;
	if (type != TYPE_CONTEXT_SPECIFIC_1 && type != TYPE_SET)
	    return ERROR_UNEXPECTED_TYPE;
    }
    if (type == TYPE_CONTEXT_SPECIFIC_1) {
	if ((result = read_CertificateRevocationLists(f, &length)) != ERROR_NONE)
	    return result;

	if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	    return result;
	if (type != TYPE_SET)
	    return ERROR_UNEXPECTED_TYPE;
    }

    if ((result = read_SingerInfos(f, &length)) != ERROR_NONE)
	return result;
    indent++;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_signedAndEnvelopedData(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    short   i = 1;  /* allow an indefinite length */

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

    *length_limit -= length;
#if VERBOSE == 1
    printf("%ssignedAndEnvelopedData (OBJECT_IDENTIFIER)\n", indent);
#endif

    /* read signedAndEnvelopedData */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    indent--;
    if ((result = read_SignedAndEnvelopedData(f, &length)) != ERROR_NONE)
	return result;
    indent++;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_DigestedData(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    short   i = 1;  /* allow an indefinite length */
    int	    version;
    digest_algorithm_t	digest;

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sDigestedData (SEQUENCE)\n", indent);
#endif

    /* read version */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_INTEGER)
	return ERROR_UNEXPECTED_TYPE;
#if VERBOSE == 1
    printf("%sVersion (INTEGER)\n", indent-1);
#endif
    if ((result = read_INTEGER(f, &length, &version)) != ERROR_NONE)
	return result;
    if (version != 0)
	return ERROR_UNEXPECTED_VERSION;

    /* read digestAlgorithmIdentifier */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_OBJECT_IDENTIFIER)
	return ERROR_UNEXPECTED_TYPE;
    indent--;
    if ((result = read_DigestAlgorithmIdentifier(f, &length, &digest)) != ERROR_NONE)
	return result;
#if VERBOSE == 1
    switch (digest) {
	case DIGEST_ALGORITHM_SHA1:
	    printf("%s(data hashed with SHA-1)\n", indent); break;
	case DIGEST_ALGORITHM_MD5:
	    printf("%s(data hashed with MD5)\n", indent); break;
	default:
	    printf("%s(unknown hashing algorithm)\n", indent); break;
    }
#endif

    /* read contentInfo */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_ContentInfo(f, &length)) != ERROR_NONE)
	return result;
    indent++;

    /* read digest */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_OCTET_STRING)
	return ERROR_UNEXPECTED_TYPE;
#if VERBOSE == 1
    printf("%sdigest (OCTET_STRING)\n", indent-1);
#endif
    if ((result = read_OCTET_STRING(f, &length, NULL, NULL)) != ERROR_NONE)
	return result;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_digestedData(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    short   i = 1;  /* allow an indefinite length */

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

    *length_limit -= length;
#if VERBOSE == 1
    printf("%sDigestedData (SEQUENCE)\n", indent);
#endif

    /* read digestedData */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    indent--;
    if ((result = read_DigestedData(f, &length)) != ERROR_NONE)
	return result;
    indent++;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_EncryptedData(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    short   i = 1; /* allow an indefinite length */
    int	    version;

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sEncryptedData (SEQUENCE)\n", indent);
#endif
    *length_limit -= length;

    /* read version */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_INTEGER)
	return ERROR_UNEXPECTED_TYPE;
#if VERBOSE == 1
    printf("%sVersion (INTEGER)\n", indent-1);
#endif
    if ((result = read_INTEGER(f, &length, &version)) != ERROR_NONE)
	return result;
    if (version != 0)
	return ERROR_UNEXPECTED_VERSION;

    /* read encryptedContentInfo */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    indent--;
    if ((result = read_EncryptedContentInfo(f, &length)) != ERROR_NONE)
	return result;
    indent++;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    if (length > 0)
	return ERROR_TOO_MUCH_DATA;

    return ERROR_NONE;
}

error_t read_encryptedData(FILE *f, long *length_limit)
{
    error_t result;
    type_t  type;
    long    length;
    short   i = 1;  /* allow an indefinite length */

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sencryptedData (SEQUENCE)\n", indent);
#endif
    *length_limit -= length;

    /* read encryptedData */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_SEQUENCE)
	return ERROR_UNEXPECTED_TYPE;
    indent--;
    if ((result = read_EncryptedData(f, &length)) != ERROR_NONE)
	return result;
    indent++;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

error_t read_ContentInfo(FILE *f, long *length_limit)
{
    long    length;
    error_t result;
    content_type_t  content_type;
    type_t  type;
    short   i = 1;  /* allow an indefinite length */

    if ((result = read_length(f, length_limit, &length, &i)) != ERROR_NONE)
	return result;
    if (length > *length_limit)
	return ERROR_BAD_LENGTH;

#if VERBOSE == 1
    printf("%sContentInfo (SEQUENCE)\n", indent);
#endif
    *length_limit -= length;

    /* read contentType */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_OBJECT_IDENTIFIER)
	return ERROR_UNEXPECTED_TYPE;
    indent--;
    if ((result = read_ContentType(f, &length, &content_type)) != ERROR_NONE)
	return result;
    indent++;

    /* read content */
    if (length == 0)
	return ERROR_NONE;  /* it's optional */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    indent--;
    if (type != TYPE_CONTEXT_SPECIFIC_0)
	return ERROR_UNEXPECTED_TYPE;

    switch(content_type) {
	case CONTENT_TYPE_DATA:
	    if ((result = read_data(f, &length)) != ERROR_NONE)
		return result;
	    break;

	case CONTENT_TYPE_SIGNED_DATA:
	    if ((result = read_signedData(f, &length)) != ERROR_NONE)
		return result;
	    break;

	case CONTENT_TYPE_ENVELOPED_DATA:
	    if ((result = read_envelopedData(f, &length)) != ERROR_NONE)
		return result;
	    break;

	case CONTENT_TYPE_SIGNED_AND_ENVELOPED_DATA:
	    if ((result = read_signedAndEnvelopedData(f, &length)) != ERROR_NONE)
		return result;
	    break;

	case CONTENT_TYPE_DIGESTED_DATA:
	    if ((result = read_digestedData(f, &length)) != ERROR_NONE)
		return result;
	    break;

	case CONTENT_TYPE_ENCRYPTED_DATA:
	    if ((result = read_encryptedData(f, &length)) != ERROR_NONE)
		return result;
	    break;

	default:    return ERROR_UNEXPECTED_CONTENT_TYPE;
    }
    indent++;

    if (i == 0) {
	if (length != 0)
	    return ERROR_TOO_MUCH_DATA;
	return ERROR_NONE;
    }
    *length_limit += length;

    /* have an indefinite length */
    if ((result = read_type(f, &length, &type)) != ERROR_NONE)
	return result;
    if (type != TYPE_NONE)
	return ERROR_UNEXPECTED_TYPE;
    if ((result = read_length(f, length_limit, &length, NULL)) != ERROR_NONE)
	return result;
    if (length != 0)
	return ERROR_BAD_LENGTH;

    return ERROR_NONE;
}

int main(int argc, char **argv)
{
    FILE    *f;
    long    len, i;
    type_t  type;
    error_t result;

    if (argc != 3) {
	usage(argv[0]);
	return -1;
    }

    if ((f = fopen(argv[1], "rb")) == NULL) {
	fprintf(stderr, "error opening %s for input.\n", argv[1]);
	perror(argv[0]);
	return -1;
    }
    if ((output = fopen(argv[2], "wb")) == NULL) {
	fprintf(stderr, "error opening %s for output.\n", argv[2]);
	perror(argv[0]);
	return -1;
    }

    /* get the file size */
    fseek(f, 0L, SEEK_END);
    len = ftell(f);
    rewind(f);

    if ((result = read_type(f, &len, &type)) != ERROR_NONE) {
	print_error(result);
	return -1;
    }
    if (type != TYPE_SEQUENCE) {
	fprintf(stderr, "expected a contentInfo structure to start.\n");
	return -1;
    }
    indent--;
    if ((result = read_ContentInfo(f, &len)) != ERROR_NONE) {
	print_error(result);
	fprintf(stderr, "at byte %ld of file.\n", ftell(f));
	return -1;
    }
    indent++;

    fclose(f);
    fclose(output);
}

