/*
 * 40-bit RC2 implementation, somewhat optimized for a brute force search.
 *
 * Copyright 1997, Chris Hall - Counterpane Systems
 *
 * There are five routines of interest:
 *
 *   void rc2_full_keyschedule(unsigned char key[128])
 *      Takes a 40-bit key in key[0..4] and expands it to a 1024-bit key
 *      schedule.  Note, this code will only work on a little-endian
 *      machine (such as the Intel 80x86 family) because the key schedule
 *      will be cast to a short[].
 *
 *      If you do not want to use the fast cycling method we propose with
 *      rc2_partial_keyschedule, you can simply use this routine for each
 *      key that you wish to test.
 *
 *   void rc2_partial_keyschedule(unsigned char key[128])
 *      Takes a 1024-bit key schedule and produces its successor.  This
 *      makes for really efficient rekeying since it takes only two
 *      operations (vs. 128) to compute a new key schedule.
 *
 *      We have mapped most of the cycles that occur, that is, if you take
 *      a key and keep calling rc2_partial_keyschedule() you will
 *      eventually cycle around to the original key.  Consequently, the
 *      best way to search through the keyspace is to give someone a
 *      starting point and a number of times to call
 *      rc2_partial_keyschedule().
 *
 *      This is a full list of the cycles.  The first column contains the
 *      hexidecimal equivalent of the starting key (store the key in the
 *      array in big-endian order, so the first two digits correspond to
 *      the first byte, etc.).  The second column contains the
 *      hexidecimal length of the cycle associated with the respective
 *      cycle.  For example, the third line specifies that starting with
 *      key 0x178c6b98aa will cycle after exactly 0x11 iterations.
 *
 *        fefefefefe : 0000000000000001
 *        077438d1f8 : 000000000000000e
 *        178c6b98aa : 0000000000000011
 *        024fe2247e : 0000000000000026
 *        003f0bbd43 : 00000000000002d2
 *        0002d493ab : 0000000000000313
 *        000747369b : 0000000000000bbb
 *        000ca52e69 : 0000000000000fba
 *        0004d23bb4 : 0000000000005310
 *        0001253edf : 00000000000055d2
 *        0002f7536a : 00000000000144b7
 *        0000e68e6d : 000000000002893b
 *        000017f9e9 : 0000000000042dff
 *        00000a9677 : 0000000000106852
 *        000029f70c : 00000000001252cb
 *        0000066167 : 0000000000186d24
 *        00000e58bb : 000000000021dbce
 *        00cb826aee : 00000000024e8f3f
 *        0000002d03 : 0000000007d13d5d
 *        0000003ea8 : 00000000082d5067
 *        000000058b : 000000000f3ebb1a
 *        0000000391 : 000000002b2a6aa4
 *        0000000187 : 000000003644d184
 *        000000000d : 0000000202f7c3fb
 *        0000000021 : 00000005ff4fa427
 *        0000000005 : 00000019739fc8b8
 *        0000000006 : 000000439bc6a73c
 *        0000000000 : 0000009a6af14923
 *
 *      We have split the cycles into ranges of 2^20 keys.  Contact
 *      hall@counterpane.com for a copy of the file (it is roughly 5MB).
 *      
 *   int rc2_trial_encrypt(unsigned short key[64], unsigned short p[4],
 *          unsigned short c[4])
 *      Performs a trial encryption using the 1024-bit key schedule.  If
 *      the encrypted plaintext (p[]) matches the desired ciphertext (c[]),
 *      this routine returns a non-zero value.  Otherwise it returns 0.
 *
 *   void rc2_decrypt(unsigned short key[64], unsigned short *c,
 *          unsigned short *p)
 *      Decrypts the ciphertext c and stores the result in p.
 *
 *   void rc2_encrypt(unsigned short key[64], unsigned short *p,
 *          unsigned short *c)
 *      Encrypts the plaintext p and stores the result in c.
 */

#include <memory.h> 	/* for memset() and memcmp() */
#include <stdio.h>

#include "rc2.h"

#define assert(e) \
do { \
    if (!(e)) { \
	fprintf(stderr, "assert failed on line %d.\n", __LINE__); \
	exit(-1); \
    } \
} while (0)

/* 256-entry permutation table, derived somehow from pi
 * note permute[x] = permute[x+256 (mod 512)] for all valid x so that we
 * can lookup sums which overflow an 8-bit counter.
 */

const unsigned char permute[512] = {
	217,120,249,196, 25,221,181,237, 40,233,253,121, 74,160,216,157,
	198,126, 55,131, 43,118, 83,142, 98, 76,100,136, 68,139,251,162,
	 23,154, 89,245,135,179, 79, 19, 97, 69,109,141,  9,129,125, 50,
	189,143, 64,235,134,183,123, 11,240,149, 33, 34, 92,107, 78,130,
	 84,214,101,147,206, 96,178, 28,115, 86,192, 20,167,140,241,220,
	 18,117,202, 31, 59,190,228,209, 66, 61,212, 48,163, 60,182, 38,
	111,191, 14,218, 70,105,  7, 87, 39,242, 29,155,188,148, 67,  3,
	248, 17,199,246,144,239, 62,231,  6,195,213, 47,200,102, 30,215,
	  8,232,234,222,128, 82,238,247,132,170,114,172, 53, 77,106, 42,
	150, 26,210,113, 90, 21, 73,116, 75,159,208, 94,  4, 24,164,236,
	194,224, 65,110, 15, 81,203,204, 36,145,175, 80,161,244,112, 57,
	153,124, 58,133, 35,184,180,122,252,  2, 54, 91, 37, 85,151, 49,
	 45, 93,250,152,227,138,146,174,  5,223, 41, 16,103,108,186,201,
	211,  0,230,207,225,158,168, 44, 99, 22,  1, 63, 88,226,137,169,
	 13, 56, 52, 27,171, 51,255,176,187, 72, 12, 95,185,177,205, 46,
	197,243,219, 71,229,165,156,119, 10,166, 32,104,254,127,193,173,
	217,120,249,196, 25,221,181,237, 40,233,253,121, 74,160,216,157,
	198,126, 55,131, 43,118, 83,142, 98, 76,100,136, 68,139,251,162,
	 23,154, 89,245,135,179, 79, 19, 97, 69,109,141,  9,129,125, 50,
	189,143, 64,235,134,183,123, 11,240,149, 33, 34, 92,107, 78,130,
	 84,214,101,147,206, 96,178, 28,115, 86,192, 20,167,140,241,220,
	 18,117,202, 31, 59,190,228,209, 66, 61,212, 48,163, 60,182, 38,
	111,191, 14,218, 70,105,  7, 87, 39,242, 29,155,188,148, 67,  3,
	248, 17,199,246,144,239, 62,231,  6,195,213, 47,200,102, 30,215,
	  8,232,234,222,128, 82,238,247,132,170,114,172, 53, 77,106, 42,
	150, 26,210,113, 90, 21, 73,116, 75,159,208, 94,  4, 24,164,236,
	194,224, 65,110, 15, 81,203,204, 36,145,175, 80,161,244,112, 57,
	153,124, 58,133, 35,184,180,122,252,  2, 54, 91, 37, 85,151, 49,
	 45, 93,250,152,227,138,146,174,  5,223, 41, 16,103,108,186,201,
	211,  0,230,207,225,158,168, 44, 99, 22,  1, 63, 88,226,137,169,
	 13, 56, 52, 27,171, 51,255,176,187, 72, 12, 95,185,177,205, 46,
	197,243,219, 71,229,165,156,119, 10,166, 32,104,254,127,193,173
    };

/* the inverse of permute[].  used to compute the 40-bit key that gives
 * a particular key schedule (for use when we find a successful key and
 * must report the 40-bit key).
 */

const unsigned char inverse[256] = {
	209,218,185,111,156,200,120,102,128, 44,248, 55,234,224, 98,164,
	203,113, 80, 39, 75,149,217, 32,157,  4,145,227, 71,106,126, 83,
	250, 58, 59,180,168,188, 95,104,  8,202,143, 20,215,192,239,123,
	 91,191, 47,229,226,140,186, 18,225,175,178, 84, 93, 89,118,219,
	 50,162, 88,110, 28, 41,100,243,233,150, 12,152, 25,141, 62, 38,
	171,165,133, 22, 64,189, 73,103,220, 34,148,187, 60,193,155,235,
	 69, 40, 24,216, 26, 66,125,204,251,101,142, 61,205, 42,163, 96,
	174,147,138, 72,151, 81, 21,247,  1, 11,183, 54,177, 46, 17,253,
	132, 45, 63, 19,136,179, 52, 36, 27,222,197, 29, 77, 43, 23, 49,
	116,169,198, 67,109, 57,144,190,195,176, 33,107,246, 15,213,153,
	 13,172, 31, 92,158,245,249, 76,214,223,137,228,139,255,199,170,
	231,237, 70, 37,182,  6, 94, 53,181,236,206,232,108, 48, 85, 97,
	 74,254,160,121,  3,240, 16,114,124,207, 82,166,167,238, 68,211,
	154, 87,146,208, 90,122, 65,127, 14,  0, 99,242, 79,  5,131,201,
	161,212,221,196, 86,244,210,119,129,  9,130, 51,159,  7,134,117,
	 56, 78,105,241,173, 35,115,135,112,  2,194, 30,184, 10,252,230,
    };

/*
 * rc2_full_keyschedule - Perform a key schedule for a 40-bit key.  We
 *   store the key schedule in little-endian format, so this routine
 *   will need to be rewritten for a big-endian machine.
 *
 * input:
 *   key - key[0..4] contains the 40-bit key.
 *
 * output:
 *   key - key[0..127] contains the 1024-bit key schedule.
 */

void
rc2_full_keyschedule(unsigned char key[128])
{
    register unsigned char x, i = 0, len = 0+4;

    /* expand the key to 128 bytes */
    x = key[len++];
    do {
	/* 0 <= x+key[i++] < 512, so we are safe looking up in permute[] */
	key[len++] = x = permute[x + key[i++]];
    } while (len < 128);

    /* the algorithm for "reducing" the effective key size munges the first
     * byte regardless.
     */
    key[0] = permute[key[0]];

    return;
}

/*
 * rc2_partial_keyschedule() - Compute a new key schedule based upon an
 *   existing key schedule.  This routine is based upon the fact that
 *   the key schedule for the key based upon bytes 1..5 of the current
 *   key schedule will be nearly identical.  The only different is that
 *   it will be shifted by 1 byte (bytes 2..127 will be come 1..126),
 *   byte 0 will be recomputed, and byte 127 will be recomputed.
 *
 *   This routine performs the recomputation necessary for bytes 0 and
 *   127.  It is assumed that the top-level code takes care of the shift
 *   (by incrementing a pointer maybe).
 *
 *  input:
 *    key - the original 1024-bit key schedule shifted 1 byte.
 *
 *  output:
 *    key - the new 1024-bit key schedule with bytes 0 and 127 recomputed.
 */

#define ENCRYPT_ROUND(a,b,c,d,k,i) \
do { \
    (a) += (((b) & ~(d)) | ((c) & (d))) + (k)[(i)]; \
    (a)  = ((a) << 1) | ((a) >> 15); \
 \
    (b) += (((c) & ~(a)) | ((d) & (a))) + (k)[(i)+1]; \
    (b)  = ((b) << 2) | ((b) >> 14); \
 \
    (c) += (((d) & ~(b)) | ((a) & (b))) + (k)[(i)+2]; \
    (c)  = ((c) << 3) | ((c) >> 13); \
 \
    (d) += (((a) & ~(c)) | ((b) & (c))) + (k)[(i)+3]; \
    (d)  = ((d) << 5) | ((d) >> 11); \
} while (0)

/*
 * rc2_trial_encrypt() - Performs a trial encryption using the scheduled
 *   key.
 *
 * input:
 *   key - the 1024-bit key schedule.
 *   p - the source plaintext
 *   c - the target ciphertext
 *
 * output:
 *   Returns a non-zero value if the trial encryption was successful and
 *   zero otherwise.
 */

int
rc2_trial_encrypt(unsigned short key[64], unsigned short *p,
    unsigned short *c)
{
    register unsigned short  x76, x54, x32, x10;

    x10 = p[0];
    x32 = p[1];
    x54 = p[2];
    x76 = p[3];

    ENCRYPT_ROUND(x10, x32, x54, x76, key,  0);  /* r = 0 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key,  4);  /* r = 1 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key,  8);  /* r = 2 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 12);  /* r = 3 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 16);  /* r = 4 */

    x10 += key[x76 & 63];
    x32 += key[x10 & 63];
    x54 += key[x32 & 63];
    x76 += key[x54 & 63];

    ENCRYPT_ROUND(x10, x32, x54, x76, key, 20);  /* r = 5  */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 24);  /* r = 6  */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 28);  /* r = 7  */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 32);  /* r = 8  */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 36);  /* r = 9  */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 40);  /* r = 10 */

    x10 += key[x76 & 63];
    x32 += key[x10 & 63];
    x54 += key[x32 & 63];
    x76 += key[x54 & 63];

    ENCRYPT_ROUND(x10, x32, x54, x76, key, 44);  /* r = 11 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 48);  /* r = 12 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 52);  /* r = 13 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 56);  /* r = 14 */

    x10 += ((x32 & ~x76) | (x54 & x76)) + key[60];
    if (x10 != c[0])
	return 0;
    x10  = (x10 << 1) | (x10 >> 15);

    x32 += ((x54 & ~x10) | (x76 & x10)) + key[61];
    if (x32 != c[1])
	return 0;
    x32  = (x32 << 2) | (x32 >> 14);
 
    x54 += ((x76 & ~x32) | (x10 & x32)) + key[62];
    if (x54 != c[2])
	return 0;
    x54  = (x54 << 3) | (x54 >> 13);

    x76 += ((x10 & ~x54) | (x32 & x54)) + key[63];
    if (x76 != c[3])
	return 0;

    return 1;
}

/*
 * rc2_trial_encrypt() - Performs an encryption using the scheduled key.
 *
 * input:
 *   key - the 1024-bit key schedule.
 *   p - the source plaintext
 *
 * output:
 *   Returns the resulting ciphertext in c.
 */

void rc2_encrypt(unsigned short key[64], unsigned short *p, unsigned short *c)
{
    register unsigned short  x76, x54, x32, x10;

    x10 = p[0];
    x32 = p[1];
    x54 = p[2];
    x76 = p[3];

    ENCRYPT_ROUND(x10, x32, x54, x76, key,  0);  /* r = 0 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key,  4);  /* r = 1 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key,  8);  /* r = 2 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 12);  /* r = 3 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 16);  /* r = 4 */

    x10 += key[x76 & 63];
    x32 += key[x10 & 63];
    x54 += key[x32 & 63];
    x76 += key[x54 & 63];

    ENCRYPT_ROUND(x10, x32, x54, x76, key, 20);  /* r = 5  */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 24);  /* r = 6  */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 28);  /* r = 7  */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 32);  /* r = 8  */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 36);  /* r = 9  */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 40);  /* r = 10 */

    x10 += key[x76 & 63];
    x32 += key[x10 & 63];
    x54 += key[x32 & 63];
    x76 += key[x54 & 63];

    ENCRYPT_ROUND(x10, x32, x54, x76, key, 44);  /* r = 11 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 48);  /* r = 12 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 52);  /* r = 13 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 56);  /* r = 14 */
    ENCRYPT_ROUND(x10, x32, x54, x76, key, 60);  /* r = 14 */

    c[0] = x10;
    c[1] = x32;
    c[2] = x54;
    c[3] = x76;
}

#define DECRYPT_ROUND(a,b,c,d,k,i) \
do { \
    (d)  = ((d) >> 5) | ((d) << 11); \
    (d) -= (((a) & ~(c)) | ((b) & (c))) + (k)[(i)+3]; \
\
    (c)  = ((c) >> 3) | ((c) << 13); \
    (c) -= (((d) & ~(b)) | ((a) & (b))) + (k)[(i)+2]; \
\
    (b)  = ((b) >> 2) | ((b) << 14); \
    (b) -= (((c) & ~(a)) | ((d) & (a))) + (k)[(i)+1]; \
 \
    (a)  = ((a) >> 1) | ((a) << 15); \
    (a) -= (((b) & ~(d)) | ((c) & (d))) + (k)[(i)]; \
} while (0)

/*
 * rc2_decrypt() - Decrypts the ciphertext using the scheduled key.
 *
 * input:
 *   key - the 1024-bit key schedule.
 *   c - the ciphertext
 *   p - the resulting plaintext
 *
 * output:
 *   Returns the decrypted ciphertext in p.
 */

void rc2_decrypt(unsigned short key[64], unsigned short *c, unsigned short *p)
{
    register unsigned short  x76, x54, x32, x10;

    x10 = c[0];
    x32 = c[1];
    x54 = c[2];
    x76 = c[3];

    DECRYPT_ROUND(x10, x32, x54, x76, key, 60);  /* r = 15 */
    DECRYPT_ROUND(x10, x32, x54, x76, key, 56);  /* r = 14 */
    DECRYPT_ROUND(x10, x32, x54, x76, key, 52);  /* r = 13 */
    DECRYPT_ROUND(x10, x32, x54, x76, key, 48);  /* r = 12 */
    DECRYPT_ROUND(x10, x32, x54, x76, key, 44);  /* r = 11 */

    x76 -= key[x54 & 63];
    x54 -= key[x32 & 63];
    x32 -= key[x10 & 63];
    x10 -= key[x76 & 63];

    DECRYPT_ROUND(x10, x32, x54, x76, key, 40);  /* r = 10 */
    DECRYPT_ROUND(x10, x32, x54, x76, key, 36);  /* r = 9  */
    DECRYPT_ROUND(x10, x32, x54, x76, key, 32);  /* r = 8  */
    DECRYPT_ROUND(x10, x32, x54, x76, key, 28);  /* r = 7  */
    DECRYPT_ROUND(x10, x32, x54, x76, key, 24);  /* r = 6  */
    DECRYPT_ROUND(x10, x32, x54, x76, key, 20);  /* r = 5  */

    x76 -= key[x54 & 63];
    x54 -= key[x32 & 63];
    x32 -= key[x10 & 63];
    x10 -= key[x76 & 63];

    DECRYPT_ROUND(x10, x32, x54, x76, key, 16);  /* r = 4  */
    DECRYPT_ROUND(x10, x32, x54, x76, key, 12);  /* r = 3  */
    DECRYPT_ROUND(x10, x32, x54, x76, key,  8);  /* r = 2  */
    DECRYPT_ROUND(x10, x32, x54, x76, key,  4);  /* r = 1  */
    DECRYPT_ROUND(x10, x32, x54, x76, key,  0);  /* r = 0  */

    p[0] = x10;
    p[1] = x32;
    p[2] = x54;
    p[3] = x76;
}


#ifdef RC2_TEST

void print_key(unsigned char *key)
{
    short   i;
    for (i = 0; i < 128; i+= 2) {
	printf("%02x%02x ", key[i], key[i+1]);
	if (i % 32 == 30)
	    printf("\n");
    }
    printf("\n");
}

void
usage(char *progname)
{
    fprintf(stderr, "usage: %s <plaintext> <ciphertext> <key> <trials>\n"
		    "  plaintext - 16-digit hexidecimal value.\n"
		    "  ciphertext - 16-digit hexidecimal value.\n"
		    "  key - 10-digit hexidecimal starting value.\n"
		    "  trials - number of keys to try, must be even.\n",
	progname);
}

unsigned char
hex(char c)
{
    if (c >= '0' && c <= '9')
	return c - '0';
    if (c >= 'A' && c <= 'Z')
	return c - 'A' + 10;
    if (c >= 'a' && c <= 'z')
	return c - 'a' + 10;
    return 0xFF;
}

int
main(int argc, char **argv)
{
    /*
     * odd_offset_key has 259 elements so we have a dummy slot at the end
     * and two dummy slots at the front.
     */
    unsigned char   odd_offset_key[257], even_offset_key[257];

    unsigned char   *oo, *eo, *oe, *ee, key_0[5], buffer[8];
    unsigned short  p[4], c[4];
    long    l, trials;
    short   i, j;

    if (argc != 5) {
	usage(argv[0]);
	return -1;
    }

    /* load the plaintext and ciphertext from the command-line */
    if (strlen(argv[1]) != 16) {
	fprintf(stderr, "invalid plaintext.\n");
	usage(argv[0]);
	return -1;
    }
    if (strlen(argv[2]) != 16) {
	fprintf(stderr, "invalid ciphertext.\n");
	usage(argv[0]);
	return -1;
    }
    for (i = 0; i < 8; i++)
	buffer[i] = (hex(argv[1][2*i]) << 4) | hex(argv[1][2*i+1]);
    memcpy(p, buffer, sizeof(buffer));
    for (i = 0; i < 8; i++)
	buffer[i] = (hex(argv[2][2*i]) << 4) | hex(argv[2][2*i+1]);
    memcpy(c, buffer, sizeof(buffer));

    /* undo the last rotations for the ciphertext so we can compare sooner. */
    c[0] = (c[0] << 15) | (c[0] >> 1);
    c[1] = (c[1] << 14) | (c[1] >> 2);
    c[2] = (c[2] << 13) | (c[2] >> 3);
    c[3] = (c[3] << 11) | (c[3] >> 5);

    /* load the starting key and the number of keys to try */
    if (strlen(argv[3]) != 10) {
	fprintf(stderr, "invalid starting key.\n");
	usage(argv[0]);
	return -1;
    }
    for (i = 0; i < 5; i++)
	key_0[i] = (hex(argv[3][2*i]) << 4) | hex(argv[3][2*i+1]);

    trials = atol(argv[4]);
    if (trials % 2 == 1 || trials < 0) {
	fprintf(stderr, "invalid number of trials.\n");
	usage(argv[0]);
	return -1;
    }

    /* need to have two different key schedules since two successive key 
     * key schedules have different alignments.  the two data structures
     * will make it so that the current key schedule always has proper
     * word alignment so we can cast to a (unsigned short *).
     */
    memset(odd_offset_key,  0, sizeof(odd_offset_key));
    memset(even_offset_key, 0, sizeof(even_offset_key));

    /* oo and eo point to the current key schedule.  they effectively
     * represent a sliding window through odd_offset_key and
     * even_offset_key respectively.  once oo and eo pass through oe
     * and ee respectively, then they are reset to just past the
     * beginning of their respective buffers.  as a result we must
     * always make sure that we update the bytes in the lower half
     * of the buffers so that they will be correct when we reset oo
     * and eo.
     */
    oo = odd_offset_key + 2;
    eo = even_offset_key + 2;
    oe = oo + 128;
    ee = eo + 128;

    memcpy(eo, key_0, sizeof(key_0));
    rc2_full_keyschedule(eo);
    memcpy(oo, eo+1, 127);
    rc2_partial_keyschedule(oo);

    /* perform trials until we find a key or run out of keys */
    for (l = 0, trials /= 2; l < trials; l++) {
	/* perform a trial encryption with the key in even_offset_key */
	if (rc2_trial_encrypt((unsigned short*)eo, p, c)) {
	    fprintf(stderr, "success.\n");
	    fprintf(stderr, "key = %02x%02x%02x%02x%02x.\n",
		inverse[eo[0]], eo[1], eo[2], eo[3], eo[4]);
	    return 0;
	}

	/* the next key for even_offset_key comes two iterations later in
	 * the cycle than the current one, *not* one.  so get the info on
	 * the next iteration from odd_offset_key and compute the second
	 * iteration.
	 */
	eo[0] = eo[128] = oo[127];

	/* check if we need to wrap back to the beginning of the buffer */
	if ((eo += 2) == ee)
	    eo = even_offset_key + 2;
	rc2_partial_keyschedule(eo);

	/* update in the event we wrap back to the beginning */
	eo[-1] = eo[127];

	/* perform a trial encryption with the key in odd_offset_key */
	if (rc2_trial_encrypt((unsigned short*)oo, p, c)) {
	    fprintf(stderr, "success.\n");
	    fprintf(stderr, "key = %02x%02x%02x%02x%02x.\n",
		inverse[oo[0]], oo[1], oo[2], oo[3], oo[4]);
	    return 0;
	}

	/* the next key for odd_offset_key comes two iterations later in
	 * the cycle than the current one, *not* one.  so get the info on
	 * the next iteration from even_offset_key and compute the second
	 * iteration.
	 */
	oo[0] = oo[128] = eo[127];

	/* check if we need to wrap back to the beginning of the buffer */
	if ((oo += 2) == oe)
	    oo = odd_offset_key + 2;
	rc2_partial_keyschedule(oo);

	/* update in the event we wrap back to the beginning */
	oo[-1] = oo[127];
    }

    printf("failed.\n");
    return 0;
}

#endif	/* RC2_TEST */

